from pathlib import Path
from typing import Iterator

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.validate_data import _check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validate_data import _get_parsed_graphs
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts
from test.e2e.setup_testcontainers.setup import get_containers


@pytest.fixture(scope="module")
def container_ports() -> Iterator[ExternalContainerPorts]:
    with get_containers() as metadata:
        yield metadata.ports


@pytest.fixture(scope="module")
def creds(container_ports: ExternalContainerPorts) -> ServerCredentials:
    return ServerCredentials(
        "root@example.com",
        "test",
        f"http://0.0.0.0:{container_ports.api}",
        f"http://0.0.0.0:{container_ports.ingest}",
    )


@pytest.fixture(scope="module")
def api_url(container_ports: ExternalContainerPorts) -> str:
    return f"http://0.0.0.0:{container_ports.api}"


@pytest.fixture(scope="module")
def shacl_validator(api_url: str) -> ShaclValidator:
    return ShaclValidator(api_url)


@pytest.fixture(scope="module")
def _create_projects(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/validate-data/generic/project.json"), creds)


@pytest.fixture(scope="module")
def unknown_classes_graphs(_create_projects: Iterator[None], api_url: str) -> RDFGraphs:
    file = Path("testdata/validate-data/generic/unknown_classes.xml")
    return _get_parsed_graphs(api_url, file)


def test_check_for_unknown_resource_classes(unknown_classes_graphs: RDFGraphs) -> None:
    result = _check_for_unknown_resource_classes(unknown_classes_graphs)
    assert isinstance(result, UnknownClassesInData)
    expected = {"onto:NonExisting", "unknown:ClassWithEverything", "unknownClass"}
    assert result.unknown_classes == expected


@pytest.mark.usefixtures("_create_projects")
def test_cardinality_correct(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/generic/cardinality_correct.xml")
    graphs = _get_parsed_graphs(api_url, file)
    cardinality_correct = _get_validation_result(graphs, shacl_validator, None)
    assert cardinality_correct.conforms


@pytest.mark.usefixtures("_create_projects")
def test_content_correct(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/generic/content_correct.xml")
    graphs = _get_parsed_graphs(api_url, file)
    content_correct = _get_validation_result(graphs, shacl_validator, None)
    assert content_correct.conforms


@pytest.mark.usefixtures("_create_projects")
def test_minimal_correct(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/generic/minimal_correct.xml")
    graphs = _get_parsed_graphs(api_url, file)
    minimal_correct = _get_validation_result(graphs, shacl_validator, None)
    assert minimal_correct.conforms


@pytest.mark.usefixtures("_create_projects")
def test_file_value_correct(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/generic/file_value_correct.xml")
    graphs = _get_parsed_graphs(api_url, file)
    file_value_correct = _get_validation_result(graphs, shacl_validator, None)
    assert file_value_correct.conforms


@pytest.mark.usefixtures("_create_projects")
def test_dsp_inbuilt_correct(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_correct.xml")
    graphs = _get_parsed_graphs(api_url, file)
    dsp_inbuilt_correct = _get_validation_result(graphs, shacl_validator, None)
    assert dsp_inbuilt_correct.conforms



import copy
import json
from collections.abc import Mapping
from datetime import datetime
from pathlib import Path

import regex
from loguru import logger
from lxml import etree

from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.xml_parsing.parse_and_transform import parse_and_clean_xml_file


def _check_input_parameters(
    xml_file: str,
    json_file: str,
) -> tuple[Path, Path]:
    """
    Transform the input parameters into Path objects
    and check if they are valid files.

    Args:
        xml_file: the XML file with the data to be replaced
        json_file: the JSON file with the mapping (dict) of internal IDs to IRIs

    Raises:
        InputError: if one of the files could not be found

    Returns:
        path objects of the input parameters
    """
    xml_file_as_path = Path(xml_file)
    if not xml_file_as_path.is_file():
        logger.error(f"File {xml_file} could not be found.")
        raise InputError(f"File {xml_file} could not be found.")

    json_file_as_path = Path(json_file)
    if not json_file_as_path.is_file():
        logger.error(f"File {json_file} could not be found.")
        raise InputError(f"File {json_file} could not be found.")

    return xml_file_as_path, json_file_as_path


def _parse_json_file(json_file: Path) -> dict[str, str]:
    """
    Read JSON file and parse it into a dictionary.

    Args:
        json_file: path to JSON file

    Returns:
        dictionary with the contents of the JSON file
    """
    with open(json_file, encoding="utf-8", mode="r") as file:
        mapping: dict[str, str] = json.load(file)
    return mapping


def _replace_resptrs(
    tree: etree._Element,
    mapping: Mapping[str, str],
    used_mapping_entries: set[str],
) -> tuple[etree._Element, set[str]]:
    """
    Replace the internal IDs in the `<resptr>` tags by IRIs.

    Args:
        tree: parsed XML file
        mapping: mapping of internal IDs to IRIs
        used_mapping_entries: IDs of the mapping that have been found in the XML and have been replaced

    Returns:
        a tuple of the modified copy of the XML tree, and the set of the IDs that have been replaced
    """
    modified_tree = copy.deepcopy(tree)
    resptr_xpath = "|".join([f"/knora/{x}/resptr-prop/resptr" for x in ["resource", "link", "region"]])
    resptr_elems = modified_tree.xpath(resptr_xpath)
    resptr_elems_replaced = 0
    for resptr_elem in resptr_elems:
        value_before = resptr_elem.text
        if value_after := mapping.get(value_before):
            resptr_elem.text = value_after
            resptr_elems_replaced += 1
            used_mapping_entries.add(value_before)

    logger.info(f"Replaced {resptr_elems_replaced}/{len(resptr_elems)} resptr links in the XML file")
    print(f"Replaced {resptr_elems_replaced}/{len(resptr_elems)} resptr links in the XML file")

    return modified_tree, used_mapping_entries


def _replace_salsah_links(
    tree: etree._Element,
    mapping: Mapping[str, str],
    used_mapping_entries: set[str],
) -> tuple[etree._Element, set[str]]:
    """
    Replace the internal IDs in the salsah-links of the `<text>` tags by IRIs.

    Args:
        tree: parsed XML file
        mapping: mapping of internal IDs to IRIs
        used_mapping_entries: IDs of the mapping that have been found in the XML and have been replaced

    Returns:
        a tuple of the modified copy of the XML tree, and the set of the IDs that have been replaced
    """
    modified_tree = copy.deepcopy(tree)
    salsah_xpath = "|".join([f"/knora/{x}/text-prop/text//a" for x in ["resource", "link", "region"]])
    salsah_links = [x for x in modified_tree.xpath(salsah_xpath) if x.attrib.get("class") == "salsah-link"]
    salsah_links_replaced = 0
    for salsah_link in salsah_links:
        value_before = regex.sub("IRI:|:IRI", "", salsah_link.attrib.get("href", ""))
        if value_after := mapping.get(value_before):
            salsah_link.attrib["href"] = value_after
            salsah_links_replaced += 1
            used_mapping_entries.add(value_before)

    logger.info(f"Replaced {salsah_links_replaced}/{len(salsah_links)} salsah-links in the XML file")
    print(f"Replaced {salsah_links_replaced}/{len(salsah_links)} salsah-links in the XML file")

    return modified_tree, used_mapping_entries


def _replace_ids_by_iris(
    tree: etree._Element,
    mapping: Mapping[str, str],
) -> etree._Element:
    """
    Iterate over the `<resptr>` tags and the salsah-links of the `<text>` tags,
    and replace the internal IDs by IRIs.
    If an internal ID cannot be found in the mapping, the original ID is kept.

    Args:
        tree: parsed XML file
        mapping: mapping of internal IDs to IRIs

    Returns:
        a modified copy of the XML tree
    """
    used_mapping_entries: set[str] = set()

    tree, used_mapping_entries = _replace_resptrs(
        tree=tree,
        mapping=mapping,
        used_mapping_entries=used_mapping_entries,
    )

    tree, used_mapping_entries = _replace_salsah_links(
        tree=tree,
        mapping=mapping,
        used_mapping_entries=used_mapping_entries,
    )

    logger.info(f"Used {len(used_mapping_entries)}/{len(mapping)} entries from the mapping file")
    print(f"Used {len(used_mapping_entries)}/{len(mapping)} entries from the mapping file")

    return tree


def _remove_resources_if_id_in_mapping(
    tree: etree._Element,
    mapping: Mapping[str, str],
) -> etree._Element:
    """
    Remove all resources from the XML file if their ID is in the mapping.

    Args:
        tree: parsed XML file
        mapping: mapping of internal IDs to IRIs

    Returns:
        a modified copy of the XML tree
    """
    modified_tree = copy.deepcopy(tree)
    resources = modified_tree.xpath("|".join([f"/knora/{x}" for x in ["resource", "link", "region"]]))
    resources_to_remove = [x for x in resources if x.attrib.get("id") in mapping]
    for resource in resources_to_remove:
        resource.getparent().remove(resource)

    msg = (
        f"Removed {len(resources_to_remove)}/{len(resources)} resources from the XML file, "
        "because their ID was in the mapping"
    )
    logger.info(msg)
    print(msg)

    return modified_tree


def _write_output_file(
    orig_xml_file: Path,
    tree: etree._Element,
) -> None:
    """
    Write modified XML file with replaced IDs to disk.

    Args:
        orig_xml_file: XML file that was provided as input
        tree: modified XML tree with replaced IDs
    """
    timestamp_str = datetime.now().strftime("%Y%m%d-%H%M%S")
    out_file = f"{orig_xml_file.stem}_replaced_{timestamp_str}.xml"
    et = etree.ElementTree(tree)
    et.write(out_file, pretty_print=True, xml_declaration=True, encoding="utf-8")
    logger.info(f"XML with replaced IDs was written to file {out_file}.")
    print(f"XML with replaced IDs was written to file {out_file}.")


def id2iri(
    xml_file: str,
    json_file: str,
    remove_resource_if_id_in_mapping: bool = False,
) -> bool:
    """
    Replace internal IDs of an XML file
    (`<resptr>` tags and salsah-links inside `<text>` tags)
    by IRIs provided in a mapping file.
    If an internal ID cannot be found in the mapping, the original ID is kept.
    The output is written to a new XML file named "[original name]_replaced_[timestamp].xml".

    Args:
        xml_file: the XML file with the data to be replaced
        json_file: the JSON file with the mapping (dict) of internal IDs to IRIs
        remove_resource_if_id_in_mapping: if True, remove all resources from the XML file if their ID is in the mapping

    Raises:
        BaseError: if one of the two input files is not a valid file

    Returns:
        success status
    """
    xml_file_as_path, json_file_as_path = _check_input_parameters(xml_file=xml_file, json_file=json_file)
    mapping = _parse_json_file(json_file_as_path)
    tree = parse_and_clean_xml_file(xml_file_as_path)
    tree = _replace_ids_by_iris(
        tree=tree,
        mapping=mapping,
    )
    if remove_resource_if_id_in_mapping:
        tree = _remove_resources_if_id_in_mapping(
            tree=tree,
            mapping=mapping,
        )
    _write_output_file(orig_xml_file=xml_file_as_path, tree=tree)
    return True



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from dsp_tools.utils.xml_parsing.models.text_value_validation import InconsistentTextValueEncodings
from dsp_tools.utils.xml_parsing.models.text_value_validation import TextValueData
from dsp_tools.utils.xml_parsing.models.text_value_validation import _make_msg_for_one_resource
from dsp_tools.utils.xml_parsing.models.text_value_validation import _make_msg_from_df


class TestInvalidTextValueEncodings:
    def test_get_problems_as_df(self) -> None:
        problems = InconsistentTextValueEncodings(
            [
                TextValueData("id1", ":simple", {"utf8", "xml"}),
                TextValueData("id1", ":rich", {"utf8", "xml"}),
                TextValueData("id2", ":rich", {"utf8", "xml"}),
            ]
        )
        expected_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1", "id2"],
                "Property Name": [":rich", ":simple", ":rich"],
            }
        )
        res_df = problems._get_problems_as_df()
        assert_frame_equal(res_df, expected_df)

    def test_make_msg_for_one_resource(self) -> None:
        test_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1"],
                "Property Name": [":rich", ":simple"],
            }
        )
        res = _make_msg_for_one_resource("id1", test_df)
        expected = "Resource ID: 'id1'\n    - Property Name: ':rich'\n    - Property Name: ':simple'"
        assert res == expected

    def test_make_msg_from_df(self) -> None:
        test_df = pd.DataFrame(
            {
                "Resource ID": ["id1", "id1", "id2", "id3"],
                "Property Name": [":rich", ":simple", ":rich", ":mixed"],
            }
        )
        res = _make_msg_from_df(test_df)
        expected = (
            "Resource ID: 'id1'\n"
            "    - Property Name: ':rich'\n"
            "    - Property Name: ':simple'"
            "\n----------------------------\n"
            "Resource ID: 'id2'\n"
            "    - Property Name: ':rich'"
            "\n----------------------------\n"
            "Resource ID: 'id3'\n"
            "    - Property Name: ':mixed'"
        )
        assert res == expected


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import json
import warnings
from dataclasses import dataclass
from typing import Protocol

from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_decimal


@dataclass
class GeometryShape(Protocol):
    color: str

    def to_json_string(self) -> str: ...


@dataclass
class Rectangle(GeometryShape):
    point1: GeometryPoint
    point2: GeometryPoint
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "rectangle",
            "lineWidth": self.line_width,
            "points": [self.point1.to_dict(), self.point2.to_dict()],
        }
        return json.dumps(json_dict)


@dataclass
class Polygon(GeometryShape):
    points: list[GeometryPoint]
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)
        if len(self.points) < 3:
            msg = (
                f"The region shape of the resource with the ID '{self.resource_id}' is a polygon."
                f"Polygons should have at least 3 points. If you would like to display a rectangle "
                f"we recommend ot use the designated rectangle shape."
            )
            warnings.warn(DspToolsUserWarning(msg))

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "polygon",
            "lineWidth": self.line_width,
            "points": [x.to_dict() for x in self.points],
        }
        return json.dumps(json_dict)


@dataclass
class Circle(GeometryShape):
    center: GeometryPoint
    radius: Vector
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "circle",
            "lineWidth": self.line_width,
            "points": [self.center.to_dict()],
            "radius": self.radius.to_dict(),
        }
        return json.dumps(json_dict)


@dataclass
class GeometryPoint:
    x: float
    y: float
    resource_id: str

    def __post_init__(self) -> None:
        if not all([is_decimal(self.x), is_decimal(self.y)]):
            msg = (
                f"The entered geometry points for the resource with the ID '{self.resource_id}' are not floats. "
                f"x: '{self.x}', y: '{self.y}'"
            )
            warnings.warn(DspToolsUserWarning(msg))
        else:
            info = []
            if not 0 <= float(self.x) <= 1:
                info.append(f"x: '{self.x}'")
            if not 0 <= float(self.y) <= 1:
                info.append(f"y: '{self.y}'")
            if info:
                msg = (
                    f"The geometry points must be larger/equal 0 and smaller/equal 1. "
                    f"The following points of the resource with the ID '{self.resource_id}' "
                    f"are not valid: {', '.join(info)}"
                )
                warnings.warn(DspToolsUserWarning(msg))
            self.x = float(self.x)
            self.y = float(self.y)

    def to_dict(self) -> dict[str, float]:
        return {"x": self.x, "y": self.y}


@dataclass
class Vector:
    x: float
    y: float
    resource_id: str

    def __post_init__(self) -> None:
        if not all([is_decimal(self.x), is_decimal(self.y)]):
            msg = (
                f"The radius vector for the resource with the ID '{self.resource_id}' are not floats. "
                f"x: '{self.x}', y: '{self.y}'"
            )
            warnings.warn(DspToolsUserWarning(msg))
        else:
            info = []
            if not 0 <= float(self.x) <= 1:
                info.append(f"x: '{self.x}'")
            if not 0 <= float(self.y) <= 1:
                info.append(f"y: '{self.y}'")
            if info:
                msg = (
                    f"The radius vector must be larger/equal 0 and smaller/equal 1. "
                    f"The following values of the resource with the ID '{self.resource_id}' "
                    f"are not valid: {', '.join(info)}"
                )
                warnings.warn(DspToolsUserWarning(msg))
            self.x = float(self.x)
            self.y = float(self.y)

    def to_dict(self) -> dict[str, float]:
        return {"x": self.x, "y": self.y}


def _check_warn_shape_info(color: str, line_width: float, res_id: str) -> None:
    problems = []
    line_msg = f"The line width must be a number between 1-5. The provided value is not valid: {line_width}"
    if not is_decimal(line_width):
        problems.append(line_msg)
    elif not 0 < int(line_width) <= 5:
        problems.append(line_msg)
    if not is_color(color):
        problems.append(f"The color must be a valid hexadecimal color. The provided value is not valid: {color}")
    if problems:
        warnings.warn(
            DspToolsUserWarning(
                f"Some of the shape values of the resource with the ID '{res_id}' are not valid: "
                f"\n- {'\n- '.join(problems)}"
            )
        )



from __future__ import annotations

import unicodedata
from typing import Any
from typing import Optional
from typing import TypeGuard

import pandas as pd
import regex

from dsp_tools.commands.excel2xml.propertyelement import PropertyElement


def simplify_name(value: str) -> str:
    """
    Simplifies a given value in order to use it as node name

    Args:
        value: The value to be simplified

    Returns:
        str: The simplified value
    """
    simplified_value = value.lower()

    # normalize characters (p.ex. ä becomes a)
    simplified_value = unicodedata.normalize("NFKD", simplified_value)

    # replace forward slash and whitespace with a dash
    simplified_value = regex.sub("[/\\s]+", "-", simplified_value)

    # delete all characters which are not letters, numbers or dashes
    simplified_value = regex.sub("[^A-Za-z0-9\\-]+", "", simplified_value)

    return simplified_value


def check_notna(value: Optional[Any]) -> TypeGuard[Any]:
    """
    Check a value if it is usable in the context of data archiving. A value is considered usable if it is
     - a number (integer or float, but not np.nan)
     - a boolean
     - a string with at least one Unicode letter (matching the regex ``\\p{L}``) or number, or at least one _, !, or ?
       (The strings `None`, `<NA>`, `N/A`, and `-` are considered invalid.)
     - a PropertyElement whose "value" fulfills the above criteria

    Args:
        value: any object encountered when analysing data

    Returns:
        True if the value is usable, False if it is N/A or otherwise unusable

    Examples:
        >>> check_notna(0)      == True
        >>> check_notna(False)  == True
        >>> check_notna("œ")    == True
        >>> check_notna("0")    == True
        >>> check_notna("_")    == True
        >>> check_notna("!")    == True
        >>> check_notna("?")    == True
        >>> check_notna(None)   == False
        >>> check_notna("None") == False
        >>> check_notna(<NA>)   == False
        >>> check_notna("<NA>") == False
        >>> check_notna("-")    == False
        >>> check_notna(" ")    == False
    """

    if isinstance(value, PropertyElement):
        value = value.value

    if isinstance(value, (bool, int)) or (
        isinstance(value, float) and pd.notna(value)
    ):  # necessary because isinstance(np.nan, float)
        return True
    elif isinstance(value, str):
        return bool(regex.search(r"[\p{L}\d_!?]", value, flags=regex.UNICODE)) and not bool(
            regex.search(r"^(none|<NA>|-|n/a)$", value, flags=regex.IGNORECASE)
        )
    else:
        return False



import pytest
from lxml import etree

from dsp_tools.utils.xml_parsing.get_data_deserialised import get_data_deserialised


def test_to_data_rdf(data_xml: etree._Element) -> None:
    shortcode, data = get_data_deserialised(data_xml)
    assert shortcode == "9999"
    assert len(data.resources) == 18


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.clients.legal_info_client_live import HTTP_LACKING_PERMISSIONS
from dsp_tools.clients.legal_info_client_live import LegalInfoClientLive
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.request_utils import RequestParameters

AUTH = Mock()
AUTH.get_token = Mock(return_value="tkn")


class TestPostCopyrightHolders:
    @patch("dsp_tools.clients.legal_info_client_live.log_response")
    @patch("dsp_tools.clients.legal_info_client_live.log_request")
    def test_log_request(self, log_request: Mock, log_response: Mock):  # noqa: ARG002
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        params = RequestParameters(
            method="POST",
            url="http://api.com/admin/projects/shortcode/9999/legal-info/copyright-holders",
            data={"data": ["1"]},
            timeout=60,
        )
        with patch("dsp_tools.clients.legal_info_client_live.requests.post") as post_mock:
            post_mock.return_value = Mock(status_code=200, ok=True)
            response = client._post_and_log_request("copyright-holders", ["1"])
        assert response.status_code == 200
        assert response.ok
        log_request_call_params = log_request.call_args_list[0].args[0]
        assert isinstance(log_request_call_params, RequestParameters)
        del log_request_call_params.headers
        assert log_request_call_params == params

    @patch("dsp_tools.clients.legal_info_client_live.log_response")
    @patch("dsp_tools.clients.legal_info_client_live.log_request")
    def test_post_ok(self, log_request: Mock, log_response: Mock):  # noqa: ARG002
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        params = RequestParameters(
            method="POST",
            url="http://api.com/admin/projects/shortcode/9999/legal-info/copyright-holders",
            data={"data": ["1"]},
            timeout=60,
            headers={"Content-Type": "application/json", "Authorization": "Bearer tkn"},
        )
        with patch("dsp_tools.clients.legal_info_client_live.requests.post") as post_mock:
            post_mock.return_value = Mock(status_code=200, ok=True)
            client._post_and_log_request("copyright-holders", ["1"])
            post_mock.assert_called_once_with(
                url=params.url, headers=params.headers, data=params.data_serialized, timeout=params.timeout
            )

    def test_client_post_copyright_holders_lacking_permissions(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        expected_response = Mock(status_code=HTTP_LACKING_PERMISSIONS, ok=False)
        client._post_and_log_request = Mock(return_value=expected_response)
        with pytest.raises(BadCredentialsError):
            client.post_copyright_holders(["1"])

    def test_client_post_copyright_holders_unknown_status_code(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        expected_response = Mock(status_code=404, ok=False)
        client._post_and_log_request = Mock(return_value=expected_response)
        with pytest.raises(BaseError):
            client.post_copyright_holders(["1"])



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Iterable
from typing import Protocol
from urllib.parse import quote_plus

from loguru import logger

from dsp_tools.clients.connection import Connection


@dataclass(frozen=True)
class ListNode:
    """Information about a list node."""

    node_iri: str
    node_name: str


@dataclass(frozen=True)
class List:
    """Information about a list."""

    root_iri: str
    list_name: str
    nodes: list[ListNode]


@dataclass(frozen=True)
class ProjectLists:
    """Information about the lists of a project."""

    lists: list[List]


class ListClient(Protocol):
    """Interface (protocol) for list-related requests to the DSP-API."""

    def get_list_node_id_to_iri_lookup(self) -> dict[tuple[str, str], str]:
        """Get a lookup of list node names to IRIs."""


@dataclass
class ListClientLive:
    """Client handling list-related requests to the DSP-API."""

    con: Connection
    project_iri: str
    list_info: ProjectLists | None = field(init=False, default=None)

    def get_list_node_id_to_iri_lookup(self) -> dict[tuple[str, str], str]:
        """
        Get a mapping of list node IDs to their respective IRIs.
        A list node ID is structured as follows:
        `("<list name>", "<node name>") where the list name is the node name of the root node.

        Returns:
            The mapping of list node IDs to IRIs.
        """
        if not self.list_info:
            self.list_info = _get_list_info_from_server(self.con, self.project_iri)
        return dict(_get_node_tuples(self.list_info.lists))


def _get_node_tuples(lists: list[List]) -> Iterable[tuple[tuple[str, str], str]]:
    for lst in lists:
        list_name = lst.list_name
        for node in lst.nodes:
            yield (list_name, node.node_name), node.node_iri


def _get_list_info_from_server(con: Connection, project_iri: str) -> ProjectLists:
    logger.info(f"Retrieving lists of project {project_iri}")
    list_iris = _get_list_iris_from_server(con, project_iri)
    lists = [_get_list_from_server(con, list_iri) for list_iri in list_iris]
    return ProjectLists(lists)


def _get_list_iris_from_server(con: Connection, project_iri: str) -> list[str]:
    iri = quote_plus(project_iri)
    res = con.get(f"/admin/lists?projectIri={iri}")
    lists: list[dict[str, Any]] = res["lists"]
    logger.info(f"Found {len(lists)} lists for project")
    return [lst["id"] for lst in lists]


def _get_list_from_server(con: Connection, list_iri: str) -> List:
    logger.info(f"Retrieving nodes of list {list_iri}")
    iri = quote_plus(list_iri)
    res = con.get(f"/admin/lists/{iri}")
    list_object: dict[str, Any] = res["list"]
    list_info = list_object["listinfo"]
    children: list[dict[str, Any]] = list_object["children"]
    root_iri = list_info["id"]
    # if the root node does not have a name, use the label instead
    list_name: str = list_info.get("name") or list_info["labels"][0]["value"]
    root_node = ListNode(root_iri, list_name)
    nodes = [root_node, *_children_to_nodes(children)]
    return List(root_iri, list_name, nodes)


def _children_to_nodes(children: list[dict[str, Any]]) -> list[ListNode]:
    nodes = []
    for child in children:
        node_iri = child["id"]
        node_name = child["name"]
        nodes.append(ListNode(node_iri=node_iri, node_name=node_name))
        if "children" in child:
            nodes.extend(_children_to_nodes(child["children"]))
    return nodes



from dataclasses import dataclass
from dataclasses import field
from functools import partial
from importlib.metadata import version
from typing import Any
from typing import Literal
from typing import cast

import regex
from requests import ReadTimeout
from requests import RequestException
from requests import Response
from requests import Session

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.connection import Connection
from dsp_tools.config.logger_config import WARNINGS_SAVEPATH
from dsp_tools.error.exceptions import InvalidInputError
from dsp_tools.error.exceptions import PermanentConnectionError
from dsp_tools.utils.request_utils import PostFiles
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_and_raise_timeouts
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_request_failure_and_sleep
from dsp_tools.utils.request_utils import log_response
from dsp_tools.utils.request_utils import should_retry

HTTP_OK = 200
HTTP_UNAUTHORIZED = 401


@dataclass
class ConnectionLive(Connection):
    """
    A Connection instance represents a connection to a DSP server.

    Attributes:
        server: address of the server, e.g https://api.dasch.swiss
        token: session token received by the server after login
    """

    server: str
    authenticationClient: AuthenticationClient | None = None
    session: Session = field(init=False, default=Session())
    # downtimes of server-side services -> API still processes request
    # -> retry too early has side effects (e.g. duplicated resources)
    timeout_put_post: int = field(init=False, default=30 * 60)
    timeout_get: int = field(init=False, default=20)

    def __post_init__(self) -> None:
        self.session.headers["User-Agent"] = f"DSP-TOOLS/{version('dsp-tools')}"
        if self.server.endswith("/"):
            self.server = self.server[:-1]
        if self.authenticationClient and (token := self.authenticationClient.get_token()):
            self.session.headers["Authorization"] = f"Bearer {token}"

    def logout(self) -> None:
        """
        Remove the authorization header from the connection's session.
        """
        del self.session.headers["Authorization"]

    def post(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        files: PostFiles | None = None,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP POST request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            data: payload of the HTTP request
            files: files to be uploaded, if any
            headers: headers for the HTTP request
            timeout: timeout of the HTTP request, or None if the default should be used

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        if data:
            headers = headers or {}
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json; charset=UTF-8"
        params = RequestParameters(
            "POST", self._make_url(route), timeout or self.timeout_put_post, data, headers, files
        )
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def get(
        self,
        route: str,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP GET request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            headers: headers for the HTTP request

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        params = RequestParameters("GET", self._make_url(route), self.timeout_get, headers=headers)
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def put(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        """
        Make an HTTP GET request to the server to which this connection has been established.

        Args:
            route: route that will be called on the server
            data: payload of the HTTP request
            headers: headers of the HTTP request

        Returns:
            response from server

        Raises:
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
        """
        if data:
            headers = headers or {}
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json; charset=UTF-8"
        params = RequestParameters("PUT", self._make_url(route), self.timeout_put_post, data, headers)
        response = self._try_network_action(params)
        return cast(dict[str, Any], response.json())

    def _make_url(self, route: str) -> str:
        if not route.startswith("/"):
            route = f"/{route}"
        return self.server + route

    def _try_network_action(self, params: RequestParameters) -> Response:
        """
        Try 7 times to execute an HTTP request.
        If a timeout error, a ConnectionError, or a requests.RequestException occur,
        or if the response indicates that there is a non-permanent server-side problem,
        this function waits and retries the HTTP request.
        The waiting times are 1, 2, 4, 8, 16, 32, 64 seconds.

        Args:
            params: keyword arguments for the HTTP request

        Raises:
            BadCredentialsError: if the server returns a 401 status code on the route /v2/authentication
            PermanentConnectionError: if all attempts have failed
            InvalidInputError: if the API responds with a permanent error because of invalid input data
            unexpected exceptions: if the action fails with an unexpected exception

        Returns:
            the return value of action
        """
        action = partial(self.session.request, **params.as_kwargs())
        for retry_counter in range(7):
            try:
                log_request(params, dict(self.session.headers))
                response = action()
            except (TimeoutError, ReadTimeout) as err:
                log_and_raise_timeouts(err)
            except (ConnectionError, RequestException):
                self._renew_session()
                log_request_failure_and_sleep(
                    reason="Connection Error raised", retry_counter=retry_counter, exc_info=True
                )
                continue

            log_response(response)
            if response.status_code == HTTP_OK:
                return response

            self._handle_non_ok_responses(response, retry_counter)

        # if all attempts have failed, raise error
        msg = f"Permanently unable to execute the network action. See {WARNINGS_SAVEPATH} for more information."
        raise PermanentConnectionError(msg)

    def _handle_non_ok_responses(self, response: Response, retry_counter: int) -> None:
        if should_retry(response):
            log_request_failure_and_sleep("Transient Error", retry_counter, exc_info=False)
            return None
        msg = "Permanently unable to execute the network action. "
        blame: Literal["server", "client"] = "server"
        if original_str := regex.search(r'{"knora-api:error":"dsp\.errors\.(.*)","@context', str(response.content)):
            msg += f"\n{' ' * 37}Original Message: {original_str.group(1)}\n"
            blame = "client"
        if original_str and original_str.group(1).startswith("OntologyConstraintException"):
            blame = "client"
        if original_str and original_str.group(1).startswith("NotFoundException"):
            blame = "client"
        if blame == "client":
            raise InvalidInputError(msg)
        else:
            raise PermanentConnectionError(msg)

    def _renew_session(self) -> None:
        self.session.close()
        self.session = Session()
        self.session.headers["User-Agent"] = f"DSP-TOOLS/{version('dsp-tools')}"
        if self.authenticationClient and (token := self.authenticationClient.get_token()):
            self.session.headers["Authorization"] = f"Bearer {token}"



"""
This model implements the handling of ontologies. It is to note that ResourceClasses, PropertyClasses
as well as the assignment of PropertyCLasses to the ResourceClasses (with a given cardinality)
is handled in "cooperation" with the propertyclass.py (PropertyClass) and resourceclass.py (ResourceClass
and HasProperty) modules.

CREATE:
    * Instantiate a new object of the Ontology class with all required parameters
    * Call the ``create``-method on the instance to create the ontology withing the backend

READ:
    * Instantiate a new object with ``iri``(IRI of ontology) given
    * Call the ``read``-method on the instance. Reading the ontology also reads all
      associated PropertyClasses and ResourceClasses as well as the assignments.
    * Access the information that has been provided to the instance

"""

from __future__ import annotations

import copy
from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

import regex

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.context import Context
from dsp_tools.commands.project.legacy_models.helpers import WithId
from dsp_tools.commands.project.legacy_models.model import Model
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.commands.project.legacy_models.propertyclass import PropertyClass
from dsp_tools.commands.project.legacy_models.resourceclass import ResourceClass
from dsp_tools.error.exceptions import BaseError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp


class Ontology(Model):
    ROUTE: str = "/v2/ontologies"
    METADATA: str = "/metadata/"
    ALL_LANGUAGES: str = "?allLanguages=true"

    _iri: str
    _project: str
    _name: str
    _label: str
    _comment: str
    _lastModificationDate: DateTimeStamp
    _resource_classes: list[ResourceClass]
    _property_classes: list[PropertyClass]
    _context: Context
    _skiplist: list[str]

    def __init__(
        self,
        con: Connection,
        iri: Optional[str] = None,
        project: Optional[Union[str, Project]] = None,
        name: Optional[str] = None,
        label: Optional[str] = None,
        comment: Optional[str] = None,
        lastModificationDate: Optional[Union[str, DateTimeStamp]] = None,
        resource_classes: Optional[list[ResourceClass]] = None,
        property_classes: Optional[list[PropertyClass]] = None,
        context: Context = None,
    ):
        super().__init__(con)
        self._iri = iri
        if isinstance(project, Project):
            self._project = project.iri
        else:
            self._project = project
        self._name = name
        self._label = label
        self._comment = comment
        if lastModificationDate is None:
            self._lastModificationDate = None
        elif isinstance(lastModificationDate, DateTimeStamp):
            self._lastModificationDate = lastModificationDate
        else:
            self._lastModificationDate = DateTimeStamp(lastModificationDate)
        self._resource_classes = resource_classes or []
        self._property_classes = property_classes or []
        self._context = context if context is not None else Context()
        self._skiplist = []

    @property
    def iri(self) -> str:
        return self._iri

    @property
    def project(self) -> str:
        return self._project

    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, value: str) -> None:
        self._label = str(value)
        self._changed.add("label")

    @property
    def comment(self) -> str:
        return self._comment

    @comment.setter
    def comment(self, value: str) -> None:
        self._comment = str(value)
        self._changed.add("comment")

    @property
    def lastModificationDate(self) -> DateTimeStamp:
        return self._lastModificationDate

    @lastModificationDate.setter
    def lastModificationDate(self, value: Union[str, DateTimeStamp]) -> None:
        self._lastModificationDate = DateTimeStamp(value)

    @property
    def resource_classes(self) -> list[ResourceClass]:
        return self._resource_classes

    @resource_classes.setter
    def resource_classes(self, value: list[ResourceClass]) -> None:
        self._resource_classes = value

    @property
    def property_classes(self) -> list[PropertyClass]:
        return self._property_classes

    @property_classes.setter
    def property_classes(self, value: list[PropertyClass]) -> None:
        self._property_classes = value

    @property
    def context(self) -> Context:
        return self._context

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> Ontology:
        #
        # First let's get the ID (IRI) of the ontology
        #
        iri = json_obj.get("@id")
        if iri is None:
            raise BaseError("Ontology iri is missing")

        #
        # evaluate the JSON-LD context to get the proper prefixes
        #
        context = Context(json_obj.get("@context"))
        onto_name = iri.split("/")[-2]
        context.add_context(onto_name, iri + "#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        this_onto = context.prefix_from_iri(iri + "#")

        label = json_obj.get(rdfs + ":label")
        if label is None:
            raise BaseError("Ontology label is missing")
        comment = json_obj.get(rdfs + ":comment")
        if json_obj.get(knora_api + ":attachedToProject") is None:
            raise BaseError("Ontology not attached to a project")
        if json_obj[knora_api + ":attachedToProject"].get("@id") is None:
            raise BaseError("Ontology not attached to a project")
        project = json_obj[knora_api + ":attachedToProject"]["@id"]
        tmp = json_obj.get(knora_api + ":lastModificationDate")
        if tmp is not None:
            last_modification_date = DateTimeStamp(json_obj.get(knora_api + ":lastModificationDate"))
        else:
            last_modification_date = None
        resource_classes = None
        property_classes = None
        if json_obj.get("@graph") is not None:
            resclasses_obj = list(
                filter(lambda a: a.get(knora_api + ":isResourceClass") is not None, json_obj.get("@graph"))
            )
            resource_classes = list(
                map(lambda a: ResourceClass.fromJsonObj(con=con, context=context, json_obj=a), resclasses_obj)
            )

            properties_obj = list(
                filter(lambda a: a.get(knora_api + ":isResourceProperty") is not None, json_obj.get("@graph"))
            )
            property_classes = [
                PropertyClass.fromJsonObj(con=con, context=context, json_obj=a)
                for a in properties_obj
                if WithId(a.get(knora_api + ":objectType")).to_string() != "knora-api:LinkValue"
            ]
        return cls(
            con=con,
            iri=iri,
            label=label,
            project=project,
            name=this_onto,
            comment=comment,
            lastModificationDate=last_modification_date,
            resource_classes=resource_classes,
            property_classes=property_classes,
            context=context,
        )

    @classmethod
    def __oneOntologiesFromJsonObj(cls, con: Connection, json_obj: Any, context: Context) -> Ontology:
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        if json_obj.get("@type") != owl + ":Ontology":
            return None
        iri = json_obj.get("@id")
        if iri is None:
            raise BaseError("Ontology iri is missing")
        if json_obj.get(knora_api + ":attachedToProject") is None:
            raise BaseError("Ontology not attached to a project (1)")
        if json_obj[knora_api + ":attachedToProject"].get("@id") is None:
            raise BaseError("Ontology not attached to a project (2)")
        project = json_obj[knora_api + ":attachedToProject"]["@id"]
        tmp = json_obj.get(knora_api + ":lastModificationDate")
        if tmp is not None:
            last_modification_date = DateTimeStamp(json_obj.get(knora_api + ":lastModificationDate"))
        else:
            last_modification_date = None
        label = json_obj.get(rdfs + ":label")
        if label is None:
            raise BaseError("Ontology label is missing")
        comment = json_obj.get(rdfs + ":comment")
        this_onto = iri.split("/")[-2]
        context2 = copy.deepcopy(context)
        context2.add_context(this_onto, iri)
        return cls(
            con=con,
            iri=iri,
            label=label,
            project=project,
            name=this_onto,
            comment=comment,
            lastModificationDate=last_modification_date,
            context=context2,
        )

    @classmethod
    def allOntologiesFromJsonObj(cls, con: Connection, json_obj: Any) -> list[Ontology]:
        context = Context(json_obj.get("@context"))
        ontos: list[Ontology] = []
        if json_obj.get("@graph") is not None:
            for o in json_obj["@graph"]:
                ontos.append(Ontology.__oneOntologiesFromJsonObj(con, o, context))
        else:
            onto = Ontology.__oneOntologiesFromJsonObj(con, json_obj, context)
            if onto is not None:
                ontos.append(onto)
        return ontos

    def create(self) -> Ontology:
        jsonobj = self._toJsonObj_create()
        result = self._con.post(Ontology.ROUTE, jsonobj)
        return Ontology.fromJsonObj(self._con, result)

    def _toJsonObj_create(self):
        rdfs = self._context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        knora_api = self._context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        if self._name is None:
            raise BaseError("There must be a valid name given!")
        if self._label is None:
            raise BaseError("There must be a valid label given!")
        if self._project is None:
            raise BaseError("There must be a valid project given!")
        tmp = {
            knora_api + ":ontologyName": self._name,
            knora_api + ":attachedToProject": {"@id": self._project},
            rdfs + ":label": self._label,
            "@context": self._context.toJsonObj(),
        }
        if self._comment is not None:
            tmp[rdfs + ":comment"] = self._comment
        return tmp

    def read(self) -> Ontology:
        result = self._con.get(Ontology.ROUTE + "/allentities/" + quote_plus(self._iri) + Ontology.ALL_LANGUAGES)
        return Ontology.fromJsonObj(self._con, result)

    @staticmethod
    def getAllOntologies(con: Connection) -> list[Ontology]:
        result = con.get(Ontology.ROUTE + Ontology.METADATA)
        return Ontology.allOntologiesFromJsonObj(con, result)

    @staticmethod
    def getProjectOntologies(con: Connection, project_id: str) -> list[Ontology]:
        if project_id is None:
            raise BaseError("Project ID must be defined!")
        result = con.get(Ontology.ROUTE + Ontology.METADATA + quote_plus(project_id) + Ontology.ALL_LANGUAGES)
        return Ontology.allOntologiesFromJsonObj(con, result)

    @staticmethod
    def getOntologyFromServer(con: Connection, shortcode: str, name: str) -> Ontology:
        if regex.search(r"[0-9A-F]{4}", shortcode):
            result = con.get("/ontology/" + shortcode + "/" + name + "/v2" + Ontology.ALL_LANGUAGES)
        else:
            result = con.get("/ontology/" + name + "/v2" + Ontology.ALL_LANGUAGES)
        return Ontology.fromJsonObj(con, result)

    def createDefinitionFileObj(self) -> dict[str, Any]:
        ontology = {
            "name": self.name,
            "label": self.label,
            "comment": self.comment,
            "properties": [],
            "resources": [],
        }
        if not self.comment:
            ontology.pop("comment")
        for prop in self.property_classes:
            if "knora-api:hasLinkToValue" in prop.superproperties:
                self._skiplist.append(self.name + ":" + prop.name)
                continue
            ontology["properties"].append(prop.createDefinitionFileObj(self.context, self.name))

        for res in self.resource_classes:
            ontology["resources"].append(res.createDefinitionFileObj(self.context, self.name, self._skiplist))

        return ontology
